/* SleezeBall v0.6 (c) 1999 Fredrik Rambris <boost@amiga.nu>
 * Homepage: http://boost.linux.kz/sleezeball/
 *
 * The outline is based on the rredir.c by Richard Huveneers.
 * Not very much of it left, though.
 *
 * The name came from the acronym SZB which stands for
 * Squid Zapps Banners.
 *
 * This is licensed under GNU GPL (see below) with the following exception:
 * Nils Olle Fritiof Strand or staff in his companies may not use or sell this 
 * software.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>

#define BUFFER_SIZE             (16*1024)

#ifndef CONFIG_FILE
#define CONFIG_FILE "/etc/squid/sleezeball.conf"
#endif

/* NOTE! Used to be sleezeball.conf!! */
#ifndef DEFINITION_FILE
#define DEFINITION_FILE "/etc/squid/sleezeball.definitions"
#endif

char **banners=NULL;
int sig_hup=1;

char *REDIRECT_URL=NULL;
char *LOG=NULL;

/* These two functions are taken from the program Squirm (C) 1998 Chris Foote & Wayne Piekarski */

char *get_date(void)
{
  time_t tp;
  char *ascitime;
  char *s;
  
  tp = time(NULL);
  
  ascitime = ctime(&tp);
  
  s = (char *)malloc(sizeof(char) * (strlen(ascitime)+1));
  if(s == NULL)
    exit(3); /* no use writing an error message, because this function
                will keep getting called! */
  strcpy(s, ascitime);
  s[strlen(ascitime) - 1] = '\0';
  
  return s;
}


void log( char *format, ...)
{
	FILE *log;
	char *date_str = NULL;
	char msg[BUFFER_SIZE];
	va_list args;

	if( !LOG ) return;

	va_start(args, format);
	if(vsprintf(msg, format, args) > (BUFFER_SIZE - 1))
	{
		/* string is longer than the maximum buffer we 
		   specified, so just return */
		return;
	}
	va_end(args);

	date_str = get_date();
	//fprintf(stderr, "%s:%s", date_str, msg);
    
	log = fopen(LOG, "at");
	if(log == NULL) return;

	fprintf(log, "%s:%s", date_str, msg);
    
	fflush(log);
    	free(date_str);
	fclose(log);
}

int check( char *url, char *banner )
{
	char buf[BUFFER_SIZE];
	char *s;
	strcpy( buf, banner );
	while( s=strrchr( buf, ' ' ) )
	{
		s[0]='\0';
		s++;
		if( !strstr( url, s ) ) return(0);
	}
	if( !strstr( url, buf ) ) return(0);
	return( 1 );
}

char *stripws( char *str )
{
	int start, end;
	if( !str ) return( NULL );
	for( start=0 ; str[start]==' ' ; start++ );
	if( str[start]=='\0' )
	{
		str[0]='\0';
		return( str );
	}
	for( end=strlen( str ) ; ((str[end]==' ') || (str[end]=='\0') || (str[end]=='\n') || (str[end]=='\r')) && end>-1 ; end-- ) str[end]='\0';
	if( end<1 ) return( str );
	memmove( str, str+start, end-start );
	str[ end-start+1 ]='\0';
	return( str );
}

char *strtolower( char *str )
{
	int counter;
	if( !str ) return( NULL );
	if( !str[0] ) return( str );
	counter=strlen( str );

	while( counter )
	{
		counter--;
		str[counter]=tolower( str[counter] );
	}
}

char *setstrptr( char *old, char *new )
{
	if( old )
	{
		free( old );
		old=NULL;
	}
	if( new )
	{
		if( new[0] )
		{
			if( old=malloc( strlen( new )+1 ) )
			{
				strcpy( old, new );
			}
		}
	}

	return( old );
}

void readconf( void )
{
	char buf[BUFFER_SIZE], *s;
	FILE *fp=NULL;
	if( fp=fopen( CONFIG_FILE, "r" ) )
	{
		while( fgets( buf, BUFFER_SIZE-1, fp ) )
		{
			/* First strip off comments */
			if( s=strchr( buf, '#' ) )
			{
				s[0]='\0';
			}
			/* Strip of whitespaces */
			if( s=strchr( buf, '=' ) )
			{
				s[0]='\0';
				s++;
				stripws( buf );
				stripws( s );
				strtolower( buf );

				if( !strcmp( buf, "redirect_url" ) )
				{
					REDIRECT_URL=setstrptr( REDIRECT_URL, s );
				}
				else if( !strcmp( buf, "log" ) )
				{
					LOG=setstrptr( LOG, s );
				}
			}
		}
	}	
}

void readdefines( void )
{
	char buf[BUFFER_SIZE], *s, *ptr;
	FILE *fp=NULL;
	int length_of_file=0, num_of_rows=0, counter, len;
	if( banners ) free( banners );
	banners=NULL;

	if( fp=fopen( DEFINITION_FILE, "r" ) )
	{
		while( fgets( buf, BUFFER_SIZE-1, fp ) )
		{
			length_of_file+=strlen( buf )+1;
			num_of_rows++;
		}

		if( banners=(char **)calloc( 1, ( sizeof( char * )*(num_of_rows+1) ) + length_of_file + 1 ) )
		{
			s=(char *)banners;
			s+=( sizeof( char * )*(num_of_rows+1) );
			counter=0;
			rewind( fp );
			while( counter<num_of_rows )
			{
				banners[counter]=fgets( s, length_of_file, fp );
				if( !banners[counter] ) break;


				/* Strip of whitespaces */
				stripws( s );

				/* skip empty lines and comment lines */
				if( s[0] == '\n' || s[0] == '#' ) continue;

				/* Avoid calling strlen more than once */
				len=strlen( s );

				length_of_file-=len+1;

				/* Make the definition lowercase to allow caseless match */
				strtolower( s );
				s+=strlen( s )+1;
				counter++;
			}
		}
		else goto error;
		fclose( fp );
		log( "[STATUS] Definitions read ok (%ld rows)\n", num_of_rows );
	}
	else goto error;

	return;
	error:
	log( "[STATUS] Error while reading definitions. Continuing w/o them.\n" );
	if( banners ) free( banners );
	banners=NULL;
	if( fp ) fclose( fp );
	fp=NULL;
	return;
}

void sighandler( int sig )
{
	/* In the next iteration the definitionfile will be reloaded */
	sig_hup=1;
	signal( SIGHUP, sighandler );
}

int main( int argc, char *argv[] )
{
	char buf[BUFFER_SIZE];
	char buf2[BUFFER_SIZE];
	char *s, *t;
	int tlu = 0;
	int counter=0;

	tlu=0; counter=0; s=NULL;

	/* make standard output line buffered */
	if (setvbuf(stdout, NULL, _IOLBF, 0) != 0) return 1;

	/* Install the signalhandler */
	signal( SIGHUP, sighandler );

	/* scan standard input */
	while (fgets(buf, BUFFER_SIZE, stdin) != NULL)
	{
		if( sig_hup )
		{
			readconf();
			readdefines();
			sig_hup=0;
		}
		/* check for too long urls */
		if ((s = strchr(buf, '\n')) == NULL)
		{
			tlu = 1;
			continue;
		}

		if (tlu) goto dont_redirect;

		if( !banners ) goto dont_redirect;
		if( !REDIRECT_URL ) goto dont_redirect;

		/* determine end of url */
		s[0] = '\0';	/* s has been set above */
		if( s = strchr(buf, ' ') ) s[0] = '\0';

		strtolower( buf );

		for( counter=0 ; banners[counter] ; counter++ )
		{
			if( check( buf, banners[counter] ) )
			{
				log( "[HIT] %s\n", buf );
				printf("%s", REDIRECT_URL );
				break;
			}
		}

		dont_redirect:
		tlu = 0;
		(void) printf("\n");
	}
	if( banners ) free( banners );
	if( REDIRECT_URL ) free( REDIRECT_URL );
	if( LOG ) free( LOG );
	return 0;
}
